@@ -677,7 +677,6 @@ void classic_parse(Sentence sent, Parse_Options opts)
 
 		if (NULL != ts_pruning)
 		{
-
 			free_tracon_sharing(ts_parsing);
 			ts_parsing = pack_sentence_for_parsing(sent);
 			print_time(opts, "Encoded for parsing");
@@ -727,6 +726,9 @@ void classic_parse(Sentence sent, Parse_Options opts)
 			process_linkages(sent, pex, opts);
 			if (IS_GENERATION(sent->dict))
 			    find_unused_disjuncts(sent, pex);
+#ifdef PC_DISPLAY
+			display_parse_choice(pex);
+#endif
 			free_extractor(pex);
 
 			post_process_lkgs(sent, opts);
