@@ -11,6 +11,7 @@
 /*************************************************************************/
 
 #include <time.h>
+#include <stdarg.h>
 
 #include "externs.h"
 
@@ -44,12 +45,12 @@ int getrusage(int who, struct rusage *rusage);
 #define MAX_PARSE_TIME_UNLIMITED -1
 #define MAX_MEMORY_UNLIMITED ((size_t) -1)
 
-/** returns the current usage time clock in seconds */
+/** Returns the CPU usage time, for this thread only, in seconds. */
 static double current_usage_time(void)
 {
-#if !defined(_WIN32)
+#if defined RUSAGE_THREAD
 	struct rusage u;
-	getrusage (RUSAGE_SELF, &u);
+	getrusage (RUSAGE_THREAD, &u);
 	return (u.ru_utime.tv_sec + ((double) u.ru_utime.tv_usec) / 1000000.0);
 #else
 	return ((double) clock())/CLOCKS_PER_SEC;
@@ -101,6 +102,21 @@ void resources_reset_space(Resources r)
 	r->space_when_parse_started = get_space_in_use();
 }
 
+bool resources_exhausted(Resources r)
+{
+	if (!r->timer_expired && !resources_timer_expired(r)) return false;
+
+	if (!r->timer_expired && (verbosity_level(D_USER_TIMES)))
+	{
+		prt_error("#### Timeout (%.2f seconds)\n",
+		          current_usage_time() - r->time_when_parse_started);
+	}
+	r->timer_expired = true;
+
+	return true;
+}
+
+#if 0 /* max_memory is not being set any more. */
 bool resources_exhausted(Resources r)
 {
 	if (r->timer_expired || r->memory_exhausted)
@@ -114,6 +130,7 @@ bool resources_exhausted(Resources r)
 
 	return (r->timer_expired || r->memory_exhausted);
 }
+#endif
 
 bool resources_timer_expired(Resources r)
 {
@@ -128,17 +145,21 @@ bool resources_memory_exhausted(Resources r)
 	else return (r->memory_exhausted || (get_space_in_use() > r->max_memory));
 }
 
-#define RES_COL_WIDTH sizeof("                                     ")
+#define RES_COL_WIDTH 40
 
 /** print out the cpu ticks since this was last called */
-static void resources_print_time(int verbosity_opt, Resources r, const char * s)
+GNUC_PRINTF(2,0)
+static void resources_print_time(Resources r, const char *fmt, va_list args)
 {
 	double now;
 	now = current_usage_time();
-	if (verbosity_opt >= D_USER_TIMES)
-	{
-		prt_error("++++ %-36s %7.2f seconds\n", s, now - r->when_last_called);
-	}
+
+	char s[128] = "";
+	vsnprintf(s, sizeof(s), fmt, args);
+
+	prt_error("++++ %-*s %7.2f seconds\n",
+	          RES_COL_WIDTH, s, now - r->when_last_called);
+
 	r->when_last_called = now;
 }
 
@@ -147,10 +168,10 @@ static void resources_print_total_time(int verbosity_opt, Resources r)
 {
 	double now;
 	now = current_usage_time();
-	r->cumulative_time += (now - r->time_when_parse_started) ;
+	r->cumulative_time += (now - r->time_when_parse_started);
 	if (verbosity_opt >= D_USER_BASIC)
 	{
-		prt_error("++++ %-36s %7.2f seconds (%.2f total)\n", "Time",
+		prt_error("++++ %-*s %7.2f seconds (%.2f total)\n", RES_COL_WIDTH, "Time",
 		          now - r->time_when_parse_started, r->cumulative_time);
 	}
 	r->time_when_parse_started = now;
@@ -160,14 +181,19 @@ static void resources_print_total_space(int verbosity_opt, Resources r)
 {
 	if (verbosity_opt >= D_USER_TIMES)
 	{
-		prt_error("++++ %-36s %zu bytes (%zu max)\n", "Total space",
+		prt_error("++++ %-*s %zu bytes (%zu max)\n", RES_COL_WIDTH, "Total space",
 		          get_space_in_use(), get_max_space_used());
 	}
 }
 
-void print_time(Parse_Options opts, const char * s)
+void print_time(Parse_Options opts, const char *fmt, ...)
 {
-	resources_print_time(opts->verbosity, opts->resources, s);
+	if (verbosity < D_USER_TIMES) return;
+	va_list args;
+
+	va_start(args, fmt);
+	resources_print_time(opts->resources, fmt, args);
+	va_end(args);
 }
 
 void parse_options_print_total_time(Parse_Options opts)
